import SwiftUI
import SwiftData

struct ContentView: View {
    var body: some View {
        TabView {
            Tab("Dashboard", systemImage: "waveform.path.ecg") {
                DashboardView()
            }
            Tab("Snapshots", systemImage: "clock.arrow.circlepath") {
                SnapshotsView()
            }
            Tab("Data Types", systemImage: "doc.text.magnifyingglass") {
                DataTypesView()
            }
            Tab("Settings", systemImage: "gearshape") {
                SettingsView()
            }
        }
    }
}

#Preview {
    ContentView()
    .modelContainer(for: [HealthSnapshot.self, SnapshotDelta.self, TypeDelta.self, TypeCount.self, YearlyCount.self, HealthRecord.self, TypeDistributionBin.self, DeviceProfile.self], inMemory: true)
        .environment(AppSettings())
}
