import Foundation
import SwiftData

@Model final class DeviceProfile {
    var deviceID: String = ""
    var name: String = ""
    var colorTag: String = "blue"

    init(deviceID: String) {
        self.deviceID = deviceID
    }
}
